# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/CommonPublicationData.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n9com/navercorp/scavenger/model/CommonPublicationData.proto\"\xe3\x01\n\x15\x43ommonPublicationData\x12\x0f\n\x07\x61pi_key\x18\x01 \x01(\t\x12\x10\n\x08\x61pp_name\x18\x02 \x01(\t\x12\x13\n\x0b\x61pp_version\x18\x03 \x01(\t\x12\x1d\n\x15\x63ode_base_fingerprint\x18\x04 \x01(\t\x12\x13\n\x0b\x65nvironment\x18\x05 \x01(\t\x12\x10\n\x08hostname\x18\x06 \x01(\t\x12\x1d\n\x15jvm_started_at_millis\x18\x07 \x01(\x03\x12\x10\n\x08jvm_uuid\x18\x08 \x01(\t\x12\x1b\n\x13published_at_millis\x18\t \x01(\x03\x42!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.CommonPublicationData_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _COMMONPUBLICATIONDATA._serialized_start=62
  _COMMONPUBLICATIONDATA._serialized_end=289
# @@protoc_insertion_point(module_scope)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class CommonPublicationData(_message.Message):
    __slots__ = ["api_key", "app_name", "app_version", "code_base_fingerprint", "environment", "hostname", "jvm_started_at_millis", "jvm_uuid", "published_at_millis"]
    API_KEY_FIELD_NUMBER: _ClassVar[int]
    APP_NAME_FIELD_NUMBER: _ClassVar[int]
    APP_VERSION_FIELD_NUMBER: _ClassVar[int]
    CODE_BASE_FINGERPRINT_FIELD_NUMBER: _ClassVar[int]
    ENVIRONMENT_FIELD_NUMBER: _ClassVar[int]
    HOSTNAME_FIELD_NUMBER: _ClassVar[int]
    JVM_STARTED_AT_MILLIS_FIELD_NUMBER: _ClassVar[int]
    JVM_UUID_FIELD_NUMBER: _ClassVar[int]
    PUBLISHED_AT_MILLIS_FIELD_NUMBER: _ClassVar[int]
    api_key: str
    app_name: str
    app_version: str
    code_base_fingerprint: str
    environment: str
    hostname: str
    jvm_started_at_millis: int
    jvm_uuid: str
    published_at_millis: int
    def __init__(self, api_key: _Optional[str] = ..., app_name: _Optional[str] = ..., app_version: _Optional[str] = ..., code_base_fingerprint: _Optional[str] = ..., environment: _Optional[str] = ..., hostname: _Optional[str] = ..., jvm_started_at_millis: _Optional[int] = ..., jvm_uuid: _Optional[str] = ..., published_at_millis: _Optional[int] = ...) -> None: ...



import logging

from scavenger.config import Config
from scavenger.internal.banner import Banner
from scavenger.internal.client import Client
from scavenger.internal.invocation_registry import InvocationRegistry
from scavenger.internal.patch import Patcher
from scavenger.internal.scan import CodeBaseScanner
from scavenger.internal.scheduler import Scheduler
from scavenger.model.CodeBasePublication_pb2 import CodeBasePublication

logger = logging.getLogger(__name__)


def start(config: Config):
    return Agent(config).start()


class Agent:
    def __init__(self, config: Config):
        self.config = config

        invocation_registry = InvocationRegistry()
        self.codebase_scanner = CodeBaseScanner(config.codebase, config.packages, config.exclude_packages, config.decorators, config.exclude_init)
        self.patcher = Patcher(
            packages=self.config.packages,
            exclude_packages=self.config.exclude_packages,
            decorators=self.config.decorators,
            exclude_init=self.config.exclude_init,
            invocation_registry=invocation_registry
        )
        self.scheduler = Scheduler(
            config=self.config,
            client=Client(self.config),
            codebase_scanner=self.codebase_scanner,
            invocation_registry=invocation_registry
        )

    def start(self):
        Banner(self.config).print()

        codebase = None
        if not self.config.async_codebase_scan_mode:
            codebase = self.codebase_scanner.scan()

        self.patcher.patch()
        if codebase is not None:
            self.scheduler.code_base_publication = CodeBasePublication(
                common_data=self.config.build_common_publication_data(codebase.get_fingerprint(self.config)),
                entry=[function.to_codebase_entry() for function in codebase.functions]
            )
        self.scheduler.start()
        return self.scheduler

    def shutdown(self):
        self.scheduler.shutdown()



import hashlib
import time
from typing import List


def current_milli_time():
    return round(time.time() * 1000)


def md5(str_):
    return hashlib.md5(bytes(str_, "utf-8")).hexdigest()


def remove_prefix(str_: str, prefix: str):
    if str_.startswith(prefix):
        return str_[len(prefix):]
    return str_


def remove_suffix(str_: str, suffix: str):
    if str_.endswith(suffix):
        return str_[0:len(str_) - len(suffix)]
    return str_


def filter_by_exclude_packages(target: str, exclude_packages: List[str]):
    for exclude_package in exclude_packages:
        if target.startswith(exclude_package):
            return True
    return False



from scavenger.model import CommonPublicationData_pb2 as _CommonPublicationData_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class InvocationDataPublication(_message.Message):
    __slots__ = ["common_data", "entry", "recording_interval_started_at_millis"]
    class InvocationDataEntry(_message.Message):
        __slots__ = ["hash"]
        HASH_FIELD_NUMBER: _ClassVar[int]
        hash: str
        def __init__(self, hash: _Optional[str] = ...) -> None: ...
    COMMON_DATA_FIELD_NUMBER: _ClassVar[int]
    ENTRY_FIELD_NUMBER: _ClassVar[int]
    RECORDING_INTERVAL_STARTED_AT_MILLIS_FIELD_NUMBER: _ClassVar[int]
    common_data: _CommonPublicationData_pb2.CommonPublicationData
    entry: _containers.RepeatedCompositeFieldContainer[InvocationDataPublication.InvocationDataEntry]
    recording_interval_started_at_millis: int
    def __init__(self, common_data: _Optional[_Union[_CommonPublicationData_pb2.CommonPublicationData, _Mapping]] = ..., entry: _Optional[_Iterable[_Union[InvocationDataPublication.InvocationDataEntry, _Mapping]]] = ..., recording_interval_started_at_millis: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/InvocationDataPublication.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from scavenger.model import CommonPublicationData_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_CommonPublicationData__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n=com/navercorp/scavenger/model/InvocationDataPublication.proto\x1a\x39\x63om/navercorp/scavenger/model/CommonPublicationData.proto\"\xda\x01\n\x19InvocationDataPublication\x12+\n\x0b\x63ommon_data\x18\x01 \x01(\x0b\x32\x16.CommonPublicationData\x12=\n\x05\x65ntry\x18\x02 \x03(\x0b\x32..InvocationDataPublication.InvocationDataEntry\x12,\n$recording_interval_started_at_millis\x18\x03 \x01(\x03\x1a#\n\x13InvocationDataEntry\x12\x0c\n\x04hash\x18\x01 \x01(\tB!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.InvocationDataPublication_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _INVOCATIONDATAPUBLICATION._serialized_start=125
  _INVOCATIONDATAPUBLICATION._serialized_end=343
  _INVOCATIONDATAPUBLICATION_INVOCATIONDATAENTRY._serialized_start=308
  _INVOCATIONDATAPUBLICATION_INVOCATIONDATAENTRY._serialized_end=343
# @@protoc_insertion_point(module_scope)



from scavenger.model import GetConfig_pb2 as _GetConfig_pb2
from scavenger.model import CodeBasePublication_pb2 as _CodeBasePublication_pb2
from scavenger.model import InvocationDataPublication_pb2 as _InvocationDataPublication_pb2
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class PublicationResponse(_message.Message):
    __slots__ = ["status"]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    status: str
    def __init__(self, status: _Optional[str] = ...) -> None: ...



import hashlib
import math
import time
from dataclasses import dataclass
from typing import List

from scavenger.config import Config
from scavenger.internal.util import md5
from scavenger.model.CodeBasePublication_pb2 import CodeBasePublication


@dataclass
class Function:
    name: str
    declaring_type: str
    signature: str
    parameter_types: str
    package_name: str

    def to_codebase_entry(self) -> CodeBasePublication.CodeBaseEntry:
        return CodeBasePublication.CodeBaseEntry(
            declaring_type=self.declaring_type,
            method_name=self.name,
            modifiers="public",
            package_name=self.package_name,
            parameter_types=self.parameter_types,
            signature=self.signature,
            signature_hash=md5(self.signature),
            visibility="public"
        )


@dataclass
class Codebase:
    functions: List[Function]

    def get_fingerprint(self, config: Config, sort: bool = False):
        m = hashlib.sha256()
        m.update(bytes(str(config.codebase), 'utf-8'))
        m.update(bytes(str(config.packages), 'utf-8'))
        m.update(bytes(str(config.exclude_packages), 'utf-8'))
        m.update(bytes(str(config.exclude_init), 'utf-8'))
        m.update(len(self.functions).to_bytes(1, 'big'))
        functions = sorted(self.functions, key=lambda x: x.name) if sort else self.functions
        for function in functions:
            m.update(bytes(function.signature, 'utf-8'))

        return m.hexdigest()


class SchedulerState:
    name: str
    interval_seconds: int
    retry_interval_seconds: int
    retry_interval_factor: int
    num_failures: int
    next_event_at_seconds: float
    clock: int

    def __init__(self, name):
        self.name = name

    def initialize(self, interval_seconds, retry_interval_seconds):
        self.interval_seconds = interval_seconds
        self.retry_interval_seconds = retry_interval_seconds
        self.next_event_at_seconds = 0
        self.reset_retry_counter()
        return self

    def reset_retry_counter(self):
        self.num_failures = 0
        self.retry_interval_factor = 1

    def update_intervals(self, interval_seconds, retry_interval_seconds):
        if self.next_event_at_seconds != 0:
            if interval_seconds < self.interval_seconds and self.retry_interval_factor == 1:
                self.next_event_at_seconds = time.time() + interval_seconds
            elif retry_interval_seconds < self.retry_interval_seconds and self.retry_interval_factor > 1:
                self.next_event_at_seconds = time.time() + retry_interval_seconds * self.retry_interval_factor

        self.interval_seconds = interval_seconds
        self.retry_interval_seconds = retry_interval_seconds

    def schedule_next(self):
        self.next_event_at_seconds = time.time() + self.interval_seconds
        self.reset_retry_counter()

    def schedule_now(self):
        self.next_event_at_seconds = 0

    def schedule_retry(self):
        back_off_limit = 5

        if self.num_failures < back_off_limit:
            self.retry_interval_factor = 1
        else:
            self.retry_interval_factor = int(math.pow(2, min(self.num_failures - back_off_limit + 1, 4)))

        self.next_event_at_seconds = time.time() + self.retry_interval_seconds * self.retry_interval_factor
        self.num_failures += 1

    def is_due_time(self):
        return time.time() >= self.next_event_at_seconds



import grpc
import requests

from scavenger.config import Config
from scavenger.internal.constant import PROCESS_UUID
from scavenger.model.CodeBasePublication_pb2 import CodeBasePublication
from scavenger.model.GetConfig_pb2 import GetConfigRequest, GetConfigResponse
from scavenger.model.GrpcAgentService_pb2_grpc import GrpcAgentServiceStub
from scavenger.model.InvocationDataPublication_pb2 import InvocationDataPublication


class Client:

    def __init__(self, config: Config):
        response = requests.get(f"{config.server_url}/javaagent/v5/initConfig?licenseKey={config.api_key}",
                                timeout=(config.http_connect_timeout_seconds, config.http_read_timeout_seconds))
        collector_url = response.json()["collectorUrl"]
        if ":" not in collector_url:
            self.channel = grpc.insecure_channel(f"{collector_url}:{80}")
        else:
            self.channel = grpc.insecure_channel(collector_url)

        self.grpc_agent_service = GrpcAgentServiceStub(self.channel)
        self.poll_config_request = GetConfigRequest(
            jvm_uuid=PROCESS_UUID,
            api_key=config.api_key
        )

    def send_codebase_publication(self, codebase_publication: CodeBasePublication):
        self.grpc_agent_service.SendCodeBasePublication(codebase_publication)

    def send_invocation_data_publication(self, invocation_data_publication: InvocationDataPublication):
        self.grpc_agent_service.SendInvocationDataPublication(invocation_data_publication)

    def poll_config(self) -> GetConfigResponse:
        return self.grpc_agent_service.PollConfig(self.poll_config_request)



from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Optional as _Optional

DESCRIPTOR: _descriptor.FileDescriptor

class GetConfigRequest(_message.Message):
    __slots__ = ["api_key", "jvm_uuid"]
    API_KEY_FIELD_NUMBER: _ClassVar[int]
    JVM_UUID_FIELD_NUMBER: _ClassVar[int]
    api_key: str
    jvm_uuid: str
    def __init__(self, api_key: _Optional[str] = ..., jvm_uuid: _Optional[str] = ...) -> None: ...

class GetConfigResponse(_message.Message):
    __slots__ = ["code_base_publisher_check_interval_seconds", "code_base_publisher_retry_interval_seconds", "config_poll_interval_seconds", "config_poll_retry_interval_seconds", "invocation_data_publisher_interval_seconds", "invocation_data_publisher_retry_interval_seconds"]
    CODE_BASE_PUBLISHER_CHECK_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    CODE_BASE_PUBLISHER_RETRY_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_POLL_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    CONFIG_POLL_RETRY_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_DATA_PUBLISHER_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    INVOCATION_DATA_PUBLISHER_RETRY_INTERVAL_SECONDS_FIELD_NUMBER: _ClassVar[int]
    code_base_publisher_check_interval_seconds: int
    code_base_publisher_retry_interval_seconds: int
    config_poll_interval_seconds: int
    config_poll_retry_interval_seconds: int
    invocation_data_publisher_interval_seconds: int
    invocation_data_publisher_retry_interval_seconds: int
    def __init__(self, config_poll_interval_seconds: _Optional[int] = ..., config_poll_retry_interval_seconds: _Optional[int] = ..., code_base_publisher_check_interval_seconds: _Optional[int] = ..., code_base_publisher_retry_interval_seconds: _Optional[int] = ..., invocation_data_publisher_interval_seconds: _Optional[int] = ..., invocation_data_publisher_retry_interval_seconds: _Optional[int] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/CodeBasePublication.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from scavenger.model import CommonPublicationData_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_CommonPublicationData__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n7com/navercorp/scavenger/model/CodeBasePublication.proto\x1a\x39\x63om/navercorp/scavenger/model/CommonPublicationData.proto\"\xb5\x02\n\x13\x43odeBasePublication\x12+\n\x0b\x63ommon_data\x18\x01 \x01(\x0b\x32\x16.CommonPublicationData\x12\x31\n\x05\x65ntry\x18\x02 \x03(\x0b\x32\".CodeBasePublication.CodeBaseEntry\x1a\xbd\x01\n\rCodeBaseEntry\x12\x16\n\x0e\x64\x65\x63laring_type\x18\x01 \x01(\t\x12\x12\n\nvisibility\x18\x02 \x01(\t\x12\x11\n\tsignature\x18\x03 \x01(\t\x12\x13\n\x0bmethod_name\x18\x04 \x01(\t\x12\x11\n\tmodifiers\x18\x05 \x01(\t\x12\x14\n\x0cpackage_name\x18\x06 \x01(\t\x12\x17\n\x0fparameter_types\x18\x07 \x01(\t\x12\x16\n\x0esignature_hash\x18\x08 \x01(\tB!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.CodeBasePublication_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _CODEBASEPUBLICATION._serialized_start=119
  _CODEBASEPUBLICATION._serialized_end=428
  _CODEBASEPUBLICATION_CODEBASEENTRY._serialized_start=239
  _CODEBASEPUBLICATION_CODEBASEENTRY._serialized_end=428
# @@protoc_insertion_point(module_scope)



import inspect
import logging
import sys
from importlib.machinery import PathFinder, SourceFileLoader, ModuleSpec
from typing import Optional, List, Dict

from scavenger.internal.invocation_registry import InvocationRegistry
from scavenger.internal.util import md5, filter_by_exclude_packages


class Finder(PathFinder):
    invocation_registry: InvocationRegistry
    packages: List[str]
    exclude_packages: List[str]

    def __init__(self, packages: List[str], exclude_packages: List[str], decorators: List[str], exclude_init: bool,
                 invocation_registry: InvocationRegistry):
        self.packages = packages
        self.exclude_packages = exclude_packages
        self.decorators = decorators
        self.exclude_init = exclude_init
        self.invocation_registry = invocation_registry

    def find_spec(self, fullname, path=None, target=None) -> ModuleSpec:
        try:
            spec = super().find_spec(fullname, path, target)

            if spec and spec.loader and isinstance(spec.loader, SourceFileLoader) and self.patch_required(fullname):
                loader = ScavengerSourceFileLoader(fullname, spec.origin, self.invocation_registry, self.decorators, self.exclude_init)
                spec.loader = loader
                return spec
        except Exception as e:
            logging.warning("Creating custom loader is failed. ", e)

    def patch_required(self, fullname) -> bool:
        if filter_by_exclude_packages(fullname, self.exclude_packages):
            return False

        return 0 < sum(1 for module_name in self.packages if
                       fullname.startswith(module_name))

    @staticmethod
    def filter_by_exclude_packages(target, exclude_packages):
        for exclude_package in exclude_packages:
            if target.startswith(exclude_package):
                return True


class ScavengerSourceFileLoader(SourceFileLoader):
    invocation_registry: InvocationRegistry

    def __init__(self, fullname, origin, invocation_registry, decorators, exclude_init):
        super().__init__(fullname, origin)
        self.invocation_registry = invocation_registry
        self.decorators = decorators
        self.exclude_init = exclude_init

    def exec_module(self, module):
        super().exec_module(module)
        try:
            self.patch_recursively(module)
        except Exception as e:
            logging.warning("Scavenger function patching is Failed. ", e)

    def patch_recursively(self, obj):
        for key, value in inspect.getmembers(obj):
            if inspect.isfunction(value) or inspect.ismethod(value):
                if not self.is_target_module(value):
                    continue

                if isinstance(inspect.getattr_static(obj, key), classmethod):
                    value = value.__func__

                if self.exclude_init and key == '__init__':
                    continue

                if self.decorators and not self.has_decorator(self.get_decorators(value)):
                    continue

                signature = f"{value.__module__}.{value.__qualname__}{inspect.signature(value)}"

                if isinstance(inspect.getattr_static(obj, key), classmethod):
                    setattr(obj, key, classmethod(self.patch_class_method(value, signature, self.invocation_registry)))
                elif isinstance(inspect.getattr_static(obj, key), staticmethod):
                    setattr(obj, key, staticmethod(self.patch(value, signature, self.invocation_registry)))
                else:
                    setattr(obj, key, self.patch(value, signature, self.invocation_registry))

            elif inspect.isclass(value) and key != "__class__":
                if not self.is_target_module(value):
                    continue

                self.patch_recursively(value)

    def is_target_module(self, child: type) -> bool:
        return child.__module__.startswith(self.name)

    def has_decorator(self, decorators):
        return sum(1 for decorator in decorators if decorator in self.decorators) >= 1

    @staticmethod
    def patch_class_method(function, signature, invocation_registry):
        def wrapper(cls, *args, **kwargs):
            invocation_registry.register(md5(signature))
            return function(cls, *args, **kwargs)

        return wrapper

    @staticmethod
    def patch(function, signature, invocation_registry):
        def wrapper(*args, **kwargs):
            invocation_registry.register(md5(signature))
            return function(*args, **kwargs)

        return wrapper

    @staticmethod
    def get_decorators(func):
        try:
            source = inspect.getsource(func)
        except TypeError as e:
            logging.warning(f"Function inspection error : {func}")
            return []
        index = source.find("def ")
        return [
            line.strip().split()[0].split("(")[0]
            for line in source[:index].strip().splitlines()
            if line.strip()[0] == "@"
        ]


class Patcher:
    _finder: Optional[Finder]
    packages: List[str]
    exclude_packages: List[str]
    store: Dict[str, int]

    def __init__(self, packages: List[str], exclude_packages: List[str], decorators: List[str], exclude_init: bool,
                 invocation_registry: InvocationRegistry):
        self._finder = None
        self.packages = packages
        self.exclude_packages = exclude_packages
        self.decorators = decorators
        self.exclude_init = exclude_init
        self.invocation_registry = invocation_registry

    def patch(self):
        finder = Finder(self.packages, self.exclude_packages, self.decorators, self.exclude_init, self.invocation_registry)
        sys.meta_path.insert(0, finder)
        self._finder = finder

    def unpatch(self):
        sys.meta_path.remove(self._finder)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from scavenger.model import CodeBasePublication_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_CodeBasePublication__pb2
from scavenger.model import GetConfig_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2
from scavenger.model import GrpcAgentService_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2
from scavenger.model import InvocationDataPublication_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_InvocationDataPublication__pb2


class GrpcAgentServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.PollConfig = channel.unary_unary(
                '/GrpcAgentService/PollConfig',
                request_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigRequest.SerializeToString,
                response_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigResponse.FromString,
                )
        self.SendCodeBasePublication = channel.unary_unary(
                '/GrpcAgentService/SendCodeBasePublication',
                request_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_CodeBasePublication__pb2.CodeBasePublication.SerializeToString,
                response_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.FromString,
                )
        self.SendInvocationDataPublication = channel.unary_unary(
                '/GrpcAgentService/SendInvocationDataPublication',
                request_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_InvocationDataPublication__pb2.InvocationDataPublication.SerializeToString,
                response_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.FromString,
                )


class GrpcAgentServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def PollConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendCodeBasePublication(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SendInvocationDataPublication(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GrpcAgentServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'PollConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.PollConfig,
                    request_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigRequest.FromString,
                    response_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigResponse.SerializeToString,
            ),
            'SendCodeBasePublication': grpc.unary_unary_rpc_method_handler(
                    servicer.SendCodeBasePublication,
                    request_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_CodeBasePublication__pb2.CodeBasePublication.FromString,
                    response_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.SerializeToString,
            ),
            'SendInvocationDataPublication': grpc.unary_unary_rpc_method_handler(
                    servicer.SendInvocationDataPublication,
                    request_deserializer=com_dot_navercorp_dot_scavenger_dot_model_dot_InvocationDataPublication__pb2.InvocationDataPublication.FromString,
                    response_serializer=com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'GrpcAgentService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class GrpcAgentService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def PollConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GrpcAgentService/PollConfig',
            com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigRequest.SerializeToString,
            com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2.GetConfigResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendCodeBasePublication(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GrpcAgentService/SendCodeBasePublication',
            com_dot_navercorp_dot_scavenger_dot_model_dot_CodeBasePublication__pb2.CodeBasePublication.SerializeToString,
            com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SendInvocationDataPublication(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/GrpcAgentService/SendInvocationDataPublication',
            com_dot_navercorp_dot_scavenger_dot_model_dot_InvocationDataPublication__pb2.InvocationDataPublication.SerializeToString,
            com_dot_navercorp_dot_scavenger_dot_model_dot_GrpcAgentService__pb2.PublicationResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import logging
from datetime import datetime, timedelta
from typing import Optional

from apscheduler.schedulers.background import BackgroundScheduler
from grpc import StatusCode
from grpc._channel import _InactiveRpcError

from scavenger.config import Config
from scavenger.internal.client import Client
from scavenger.internal.invocation_registry import InvocationRegistry
from scavenger.internal.model import SchedulerState
from scavenger.internal.scan import CodeBaseScanner
from scavenger.model.CodeBasePublication_pb2 import CodeBasePublication
from scavenger.model.InvocationDataPublication_pb2 import InvocationDataPublication

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.DEBUG)


class Scheduler:
    def __init__(self, config: Config, client: Client, codebase_scanner: CodeBaseScanner, invocation_registry: InvocationRegistry):
        self.config = config
        self.client = client
        self.codebase_scanner = codebase_scanner
        self.invocation_registry = invocation_registry

        self.scheduler = BackgroundScheduler(daemon=False)
        self.scheduler.add_job(self.run,
                               "interval",
                               seconds=self.config.scheduler_interval_seconds,
                               next_run_time=datetime.now() + timedelta(seconds=self.config.scheduler_initial_delay_seconds))
        self.is_code_base_published = False
        self.code_base_publication: Optional[CodeBasePublication] = None
        self.invocation_data_publication: Optional[InvocationDataPublication] = None
        self.code_base_fingerprint = None
        self.dynamic_config = None
        interval_seconds = 600
        retry_interval_seconds = 600
        self.force_interval_seconds = self.config.force_interval_seconds
        self.poll_state = SchedulerState("config_pool").initialize(interval_seconds, retry_interval_seconds)
        self.code_base_publisher_state = SchedulerState("codebase").initialize(interval_seconds, retry_interval_seconds)
        self.invocation_data_publisher_state = SchedulerState("invocation_data").initialize(interval_seconds, retry_interval_seconds)

        if self.force_interval_seconds != 0:
            self.poll_state.update_intervals(self.force_interval_seconds, self.force_interval_seconds)
            self.code_base_publisher_state.update_intervals(self.force_interval_seconds, self.force_interval_seconds)
            self.invocation_data_publisher_state.update_intervals(self.force_interval_seconds, self.force_interval_seconds)

    def start(self):
        logging.info("Scheduler is starting")
        self.scheduler.start()

    def run(self):
        logging.debug("Task is starting")
        self.poll_dynamic_config_if_needed()
        self.publish_code_base_if_needed()
        self.publish_invocation_data_if_needed()
        logging.debug("Task is done")

    def shutdown(self):
        self.scheduler.shutdown()
        if self.dynamic_config is not None:
            self.code_base_publisher_state.schedule_now()
            self.publish_code_base_if_needed()

            self.invocation_data_publisher_state.schedule_now()
            self.publish_invocation_data_if_needed()

    def poll_dynamic_config_if_needed(self):
        if self.poll_state.is_due_time():
            logging.debug("Poll dynamic config is due time.")
            # noinspection PyBroadException
            try:
                self.dynamic_config = self.client.poll_config()
                if self.force_interval_seconds == 0:
                    self.poll_state.update_intervals(
                        self.dynamic_config.config_poll_interval_seconds,
                        self.dynamic_config.config_poll_retry_interval_seconds
                    )
                    self.code_base_publisher_state.update_intervals(
                        self.dynamic_config.code_base_publisher_check_interval_seconds,
                        self.dynamic_config.code_base_publisher_retry_interval_seconds
                    )
                    self.invocation_data_publisher_state.update_intervals(
                        self.dynamic_config.invocation_data_publisher_interval_seconds,
                        self.dynamic_config.invocation_data_publisher_retry_interval_seconds
                    )
                self.poll_state.schedule_next()
            except Exception as e:
                if isinstance(e, _InactiveRpcError):
                    if e.code() == StatusCode.UNAUTHENTICATED:
                        logging.warning("Authentication is Failed.")
                        self.shutdown()
                logging.warning(e)
                self.poll_state.schedule_retry()

    def publish_code_base_if_needed(self):
        if not self.is_code_base_published and self.code_base_publisher_state.is_due_time() and self.dynamic_config is not None:
            logging.debug("Publish codebase is due time.")
            try:
                codebase = self.codebase_scanner.scan()

                if len(codebase.functions) > 100_000:
                    logging.warning("maximum methods count(100000) exceed", len(codebase.functions))
                    return

                if len(codebase.functions) == 0:
                    logging.warning("no methods are found")
                    return

                self.code_base_publication = CodeBasePublication(
                    common_data=self.config.build_common_publication_data(codebase.get_fingerprint(self.config)),
                    entry=[function.to_codebase_entry() for function in codebase.functions]
                )

                self.client.send_codebase_publication(self.code_base_publication)
                self.is_code_base_published = True
                self.code_base_fingerprint = self.code_base_publication.common_data.code_base_fingerprint
                self.code_base_publication = None
            except Exception as e:
                logging.warning(e)
                self.code_base_publisher_state.schedule_retry()

    def publish_invocation_data_if_needed(self):
        if self.invocation_data_publisher_state.is_due_time() and self.dynamic_config is not None and self.is_code_base_published:
            logging.debug("Publish invocation data is due time.")
            try:
                if self.invocation_data_publication is None:
                    invocations, recording_interval_started_at_millis = self.invocation_registry.get_invocations()
                    self.invocation_data_publication = InvocationDataPublication(
                        common_data=self.config.build_common_publication_data(self.code_base_fingerprint),
                        entry=[InvocationDataPublication.InvocationDataEntry(hash=hash_) for hash_ in invocations],
                        recording_interval_started_at_millis=recording_interval_started_at_millis
                    )
                self.client.send_invocation_data_publication(
                    self.invocation_data_publication
                )
                self.invocation_data_publication = None
                self.invocation_data_publisher_state.schedule_next()
            except Exception as e:
                logging.warning(e)
                self.invocation_data_publisher_state.schedule_retry()



import ast
import logging
import os
import time
from _ast import FunctionDef, Module, Call, Attribute, Name
from dataclasses import dataclass
from pathlib import Path
from typing import List

from scavenger.internal.model import Codebase, Function
from scavenger.internal.util import remove_suffix, filter_by_exclude_packages, remove_prefix

logger = logging.getLogger(__name__)


@dataclass
class PyFile:
    codebase_path: Path
    relative_path: Path


class CodeBaseScanner:

    def __init__(self, codebase: List[str], packages: List[str], exclude_packages: List[str], decorators: List[str], exclude_init: bool):
        self.codebase_path_list = [Path(codebase_path) for codebase_path in codebase]
        self.packages = packages
        self.exclude_packages = exclude_packages
        self.decorators = decorators
        self.exclude_init = exclude_init

    def scan(self) -> Codebase:
        logger.info("Codebase scanning is starting.")
        start: int = time.perf_counter_ns()
        functions: List[Function] = []

        for py_file in self.find_all_py_files():
            with open(py_file.codebase_path.joinpath(py_file.relative_path), "r") as r:
                root: Module = ast.parse(r.read())

            package: str = remove_suffix(str(py_file.relative_path).replace(os.sep, "."), ".py")

            function_nodes: List[FunctionDef] = self.get_all_functions_from_tree(root)

            for functionDef in function_nodes:
                functions.append(self.function_def_to_function(functionDef, package))

        logger.info(
            f"Codebase scanning is done. Found {len(functions)} functions. {int((time.perf_counter_ns() - start) / 1_000_000)}ms elapsed")

        return Codebase(functions=functions)

    def get_all_functions_from_tree(self, root: Module) -> List[FunctionDef]:
        function_nodes: List[FunctionDef] = []

        for node in ast.walk(root):
            if isinstance(node, ast.ClassDef):
                for child in ast.iter_child_nodes(node):
                    child._parent = remove_prefix(f"{getattr(node, '_parent', '')}.{node.name}", ".")
            elif isinstance(node, ast.FunctionDef):
                if self.exclude_init and node.name == '__init__':
                    continue

                if self.decorators and not self.has_decorator(self.get_decorators(node)):
                    continue

                function_nodes.append(node)
        return function_nodes

    def has_decorator(self, decorators):
        return sum(1 for decorator in decorators if decorator in self.decorators) >= 1

    @staticmethod
    def get_decorators(function_def: ast.FunctionDef):
        decorators = []
        for decorator in function_def.decorator_list:
            decorators.append(f"@{get_decorator_from_node(decorator)}")
        return decorators

    @staticmethod
    def function_def_to_function(function: FunctionDef, package: str) -> Function:
        parameter_types: List[str] = [arg.arg for arg in function.args.args]

        if function.args.vararg is not None:
            parameter_types.append(f"*{function.args.vararg.arg}")
        if function.args.kwarg is not None:
            parameter_types.append(f"**{function.args.kwarg.arg}")

        parameter_types_str = ", ".join(parameter_types)
        declaring_type = remove_suffix(f"{package}.{getattr(function, '_parent', '')}", ".")

        return Function(
            declaring_type=declaring_type,
            name=function.name,
            parameter_types=parameter_types_str,
            signature=f"{declaring_type}.{function.name}({parameter_types_str})",
            package_name=package
        )

    def find_all_py_files(self) -> List[PyFile]:
        py_files: List[PyFile] = []

        for codebase_path in self.codebase_path_list:
            exclude_packages = [str(codebase_path.joinpath(exclude_package.replace(".", os.sep))) for exclude_package in
                                self.exclude_packages]

            for package in self.packages:
                py_files += self.find_files_in_package(codebase_path, exclude_packages, package)

        return py_files

    @staticmethod
    def find_files_in_package(codebase_path, exclude_packages, package):
        py_files = []
        pattern: str = os.path.join(package.replace(".", os.sep), "**", "*.py")
        for absolute_path in codebase_path.glob(pattern):
            if not filter_by_exclude_packages(str(absolute_path), exclude_packages):
                py_files.append(PyFile(codebase_path, absolute_path.relative_to(codebase_path)))
        return py_files


def get_decorator_from_node(node):
    if isinstance(node, Name):
        return node.id
    elif isinstance(node, Call):
        return f"{get_decorator_from_node(node.func)}"
    elif isinstance(node, Attribute):
        return f"{get_decorator_from_node(node.value)}.{node.attr}"
    else:
        raise logger.warning(f"Unknown decorator type : {node}")



from scavenger.model import CommonPublicationData_pb2 as _CommonPublicationData_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class CodeBasePublication(_message.Message):
    __slots__ = ["common_data", "entry"]
    class CodeBaseEntry(_message.Message):
        __slots__ = ["declaring_type", "method_name", "modifiers", "package_name", "parameter_types", "signature", "signature_hash", "visibility"]
        DECLARING_TYPE_FIELD_NUMBER: _ClassVar[int]
        METHOD_NAME_FIELD_NUMBER: _ClassVar[int]
        MODIFIERS_FIELD_NUMBER: _ClassVar[int]
        PACKAGE_NAME_FIELD_NUMBER: _ClassVar[int]
        PARAMETER_TYPES_FIELD_NUMBER: _ClassVar[int]
        SIGNATURE_FIELD_NUMBER: _ClassVar[int]
        SIGNATURE_HASH_FIELD_NUMBER: _ClassVar[int]
        VISIBILITY_FIELD_NUMBER: _ClassVar[int]
        declaring_type: str
        method_name: str
        modifiers: str
        package_name: str
        parameter_types: str
        signature: str
        signature_hash: str
        visibility: str
        def __init__(self, declaring_type: _Optional[str] = ..., visibility: _Optional[str] = ..., signature: _Optional[str] = ..., method_name: _Optional[str] = ..., modifiers: _Optional[str] = ..., package_name: _Optional[str] = ..., parameter_types: _Optional[str] = ..., signature_hash: _Optional[str] = ...) -> None: ...
    COMMON_DATA_FIELD_NUMBER: _ClassVar[int]
    ENTRY_FIELD_NUMBER: _ClassVar[int]
    common_data: _CommonPublicationData_pb2.CommonPublicationData
    entry: _containers.RepeatedCompositeFieldContainer[CodeBasePublication.CodeBaseEntry]
    def __init__(self, common_data: _Optional[_Union[_CommonPublicationData_pb2.CommonPublicationData, _Mapping]] = ..., entry: _Optional[_Iterable[_Union[CodeBasePublication.CodeBaseEntry, _Mapping]]] = ...) -> None: ...



import dataclasses
from configparser import RawConfigParser
from dataclasses import dataclass, field
from pathlib import Path
from typing import List

from dataclass_builder import dataclass_builder

from scavenger.internal.constant import HOSTNAME, PROCESS_START_AT_MILLIS, PROCESS_UUID
from scavenger.internal.util import current_milli_time
from scavenger.model.CommonPublicationData_pb2 import CommonPublicationData


@dataclass
class Config:
    server_url: str
    api_key: str
    codebase: List[str]
    packages: List[str]
    exclude_packages: List[str] = field(default_factory=list)
    decorators: List[str] = field(default_factory=list)
    exclude_init: bool = False
    app_name: str = "missing-appName"
    app_version: str = "unspecified"
    environment: str = "<default>"
    async_codebase_scan_mode: bool = False
    debug_mode: bool = False
    scheduler_initial_delay_seconds: int = 10
    scheduler_interval_seconds: int = 10
    force_interval_seconds: int = 0
    http_connect_timeout_seconds: int = 10
    http_read_timeout_seconds: int = 10

    @staticmethod
    def load_config(file: str = "scavenger.conf"):
        file = Path(file)
        if not file.exists():
            raise FileNotFoundError(f"{file.name} not found")
        with open(file, 'r') as f:
            config_string = '[dummy]\n' + f.read()
        config_from_file = RawConfigParser()
        config_from_file.optionxform = lambda x: x
        config_from_file.read_string(config_string)
        config_from_file = config_from_file["dummy"]

        config_builder = dataclass_builder(Config)()
        aliases_processors = Config.get_aliases_processors()

        for key in config_from_file:
            if key in aliases_processors:
                target_key, processor = aliases_processors[key]
                setattr(config_builder, target_key, processor(config_from_file[key]))

        return config_builder.build()

    @staticmethod
    def get_aliases_processors():
        aliases_processors = {}
        for field_ in dataclasses.fields(Config):
            key = field_.name
            processor = Config.get_processor(field_)
            aliases_processors[key] = (key, processor)

        aliases_processors["scheduler_initial_delay_millis"] = (aliases_processors["scheduler_initial_delay_seconds"][0],
                                                                lambda x: int(aliases_processors["scheduler_initial_delay_seconds"][1](x) / 1000))
        aliases_processors["scheduler_interval_millis"] = (aliases_processors["scheduler_interval_seconds"][0],
                                                           lambda x: int(aliases_processors["scheduler_interval_seconds"][1](x) / 1000))

        # Support camel cases.
        for key, value in list(aliases_processors.items()):
            aliases_processors[(Config.to_camel_case(key))] = value

        return aliases_processors

    @staticmethod
    def get_processor(field_):
        if field_.type == int:
            return lambda x: int(x)
        elif field_.type == bool:
            return lambda x: bool(x)
        elif field_.type == List[str]:
            return lambda x: [attr.strip() for attr in x.split(",")]
        else:
            return lambda x: x

    @staticmethod
    def to_camel_case(snake_str):
        components = snake_str.split('_')

        return components[0] + ''.join(x.title() for x in components[1:])

    def build_common_publication_data(self, codebase_fingerprint) -> CommonPublicationData:
        return CommonPublicationData(
            api_key=self.api_key,
            app_name=self.app_name,
            code_base_fingerprint=codebase_fingerprint,
            environment=self.environment,
            hostname=HOSTNAME,
            jvm_started_at_millis=PROCESS_START_AT_MILLIS,
            jvm_uuid=PROCESS_UUID,
            app_version=self.app_version,
            published_at_millis=current_milli_time()
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/InitConfig.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.com/navercorp/scavenger/model/InitConfig.proto\"+\n\x12InitConfigResponse\x12\x15\n\rcollector_url\x18\x01 \x01(\tB!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.InitConfig_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _INITCONFIGRESPONSE._serialized_start=50
  _INITCONFIGRESPONSE._serialized_end=93
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/GetConfig.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-com/navercorp/scavenger/model/GetConfig.proto\"5\n\x10GetConfigRequest\x12\x0f\n\x07\x61pi_key\x18\x01 \x01(\t\x12\x10\n\x08jvm_uuid\x18\x02 \x01(\t\"\xbb\x02\n\x11GetConfigResponse\x12$\n\x1c\x63onfig_poll_interval_seconds\x18\x01 \x01(\x05\x12*\n\"config_poll_retry_interval_seconds\x18\x02 \x01(\x05\x12\x32\n*code_base_publisher_check_interval_seconds\x18\x03 \x01(\x05\x12\x32\n*code_base_publisher_retry_interval_seconds\x18\x04 \x01(\x05\x12\x32\n*invocation_data_publisher_interval_seconds\x18\x05 \x01(\x05\x12\x38\n0invocation_data_publisher_retry_interval_seconds\x18\x06 \x01(\x05\x42!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.GetConfig_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _GETCONFIGREQUEST._serialized_start=49
  _GETCONFIGREQUEST._serialized_end=102
  _GETCONFIGRESPONSE._serialized_start=105
  _GETCONFIGRESPONSE._serialized_end=420
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: com/navercorp/scavenger/model/GrpcAgentService.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from scavenger.model import GetConfig_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_GetConfig__pb2
from scavenger.model import CodeBasePublication_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_CodeBasePublication__pb2
from scavenger.model import InvocationDataPublication_pb2 as com_dot_navercorp_dot_scavenger_dot_model_dot_InvocationDataPublication__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n4com/navercorp/scavenger/model/GrpcAgentService.proto\x1a-com/navercorp/scavenger/model/GetConfig.proto\x1a\x37\x63om/navercorp/scavenger/model/CodeBasePublication.proto\x1a=com/navercorp/scavenger/model/InvocationDataPublication.proto\"%\n\x13PublicationResponse\x12\x0e\n\x06status\x18\x01 \x01(\t2\xe7\x01\n\x10GrpcAgentService\x12\x35\n\nPollConfig\x12\x11.GetConfigRequest\x1a\x12.GetConfigResponse\"\x00\x12G\n\x17SendCodeBasePublication\x12\x14.CodeBasePublication\x1a\x14.PublicationResponse\"\x00\x12S\n\x1dSendInvocationDataPublication\x12\x1a.InvocationDataPublication\x1a\x14.PublicationResponse\"\x00\x42!\n\x1d\x63om.navercorp.scavenger.modelP\x01\x62\x06proto3')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'com.navercorp.scavenger.model.GrpcAgentService_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\035com.navercorp.scavenger.modelP\001'
  _PUBLICATIONRESPONSE._serialized_start=223
  _PUBLICATIONRESPONSE._serialized_end=260
  _GRPCAGENTSERVICE._serialized_start=263
  _GRPCAGENTSERVICE._serialized_end=494
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['scavenger', 'scavenger.internal', 'scavenger.model']

package_data = \
{'': ['*']}

install_requires = \
['apscheduler>=3.10.0,<4.0.0',
 'dataclass-builder>=1.2.0,<2.0.0',
 'grpcio>=1.51.1,<2.0.0',
 'path>=16.6.0,<17.0.0',
 'protobuf>=4.21.12,<5.0.0',
 'requests>=2.28.2,<3.0.0']

setup_kwargs = {
    'name': 'scavenger-agent-python',
    'version': '0.1.2',
    'description': '',
    'long_description': '# Scavenger Agent Python (BETA)\n\n*(본 Agent는 베타 버전으로 현재 개발 환경에서만의 사용을 권장하며 상용 환경에서의 사용을 권장하지 않습니다.)* Scavenger Agent의 Python 버전으로, Agent가 아닌 [Server Component](../doc/installation.md)는 사전에 준비되어야 합니다.\n\n## 개발 가이드\n\nScavenger Agent Python은 패키지 관리자로 [Poetry](https://python-poetry.org/)를 이용합니다.\n\n- 의존성 설치\n```sh\n$ poetry install\n```\n- 테스트\n```sh\n$ poetry run python -m unittest\n```\n- 빌드\n```sh\n$ poetry build\n```\n\n자세한 내용은 [Poetry Docs](https://python-poetry.org/docs/)를 참고해주세요.\n\n## 에이전트 설치 가이드\n\n### 전제조건\n\n`Python >= 3.7`\n\n### 설치\n\n```\n$ pip install scavenger-agent-python --save\n```\n### 설정\n\n설정은 설정 파일 `scavenger.conf`을 이용한 방식과 Config Instance를 직접 생성하는 두 가지 방식을 지원합니다.\n\n* Example of `scavenger.conf`\n```py\n# scavenger.conf\napiKey=eb99ff0f-aaaa-bbbb-cccc-5d1ec81f6183\nserverUrl=http://10.106.93.41:8081\nenvironment=dev\nappName=apiserver\npackages=views\ncodebase=.\n---\n# source\nconfig = Config.load_config()\nagent = Agent(config)\n```\n* Example of `Config()`\n```py\nconfig = Config(\n    api_key="eb99ff0f-aaaa-bbbb-cccc-5d1ec81f6183",\n    server_url="http://10.106.93.41:8081",\n    environment="dev",\n    app_name="apiserver",\n    packages=["views"],\n    codebase=["."],\n)\nagent = Agent(config)\n```\n\n### 에이전트 시작\nScavenger Agent Python은 Scavenger Agent Java와 달리, Agent 시작 코드를 직접 삽입하는 방식으로 동작합니다. 수집을 시작할 함수의 모듈이 import되기 전에 Agent가 동작해야 하므로 *반드시 프로그램의 시작점에 Agent 시작 코드가 삽입되어야 함을 유의하십시오.*\n```py\nfrom scavenger import Agent, Config\n\nconfig = Config.load_config()\nagent = Agent(config)\nagent.start()\n\n### your source code\nfrom ...\n```\n\n### 한계\n* Graceful Shutdown에 대한 구현은 기존 프레임워크와의 충돌 가능성으로 인해 자동으로 지원하지 않습니다. 대신 아래와 같은 `agent.shutdown()` 함수를 지원하므로 직접 등록하여서 사용하여야 합니다.\n```py\ndef shutdown_gracefully(*args):\n    # Shutdown your application/server first.\n    agent.shutdown()\n    sys.exit(0)\n\nsignal.signal(signal.SIGTERM, shutdown_gracefully)\nagent.start()\n```\n* Scavenger Python Agent는 현재 모듈/클래스의 레퍼런스를 대체하는 방식으로 동작하므로 레퍼런스를 사용하지 않고 함수를 호출하는 경우에는 동작하지 않습니다. 따라서 아래와 같이 Decorator를 이용해 함수의 레퍼런스만 따로 프레임워크에 저장해두는 경우, `def hello_world()`에는 동작하지 않습니다. 이는 추후 함수내 코드를 삽입하는 방식으로 변경되어 지원될 예정입니다.\n```py\nfrom flask import Flask\napp = Flask(__name__)\n\n@app.route(\'/\')\ndef hello_world():\n    return \'Hello World!\'\n```\n* 코드를 직접 읽어서 메서드를 탐색해야 하므로 pyc 파일은 현재 지원되지 않습니다.\n',
    'author': 'dbgsprw',
    'author_email': 'dbgsprw@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import time
from _queue import Empty
from queue import Queue
from threading import Thread
from typing import List, Set

from scavenger.internal.util import current_milli_time


class InvocationRegistry:
    queue: Queue[str]
    invocations: List[Set[str]]
    recording_interval_started_at_millis: int
    current_index: int
    FRONT_BUFFER_INDEX = 0
    BACK_BUFFER_INDEX = 1

    def __init__(self):
        self.invocations = [set(), set()]
        self.queue = Queue()
        self.current_index = self.FRONT_BUFFER_INDEX
        self.recording_interval_started_at_millis = current_milli_time()
        self.thread = Thread(target=self._update_invocations, daemon=True)
        self.thread.start()

    def register(self, hash_: str):
        self.queue.put(hash_)

    def get_invocations(self):
        time.sleep(0.01)
        old_recording_interval_started_at_millis = self.recording_interval_started_at_millis
        old_index = self.current_index
        self._toggle_index()
        invocations = list(self.invocations[old_index])
        self.invocations[old_index].clear()
        return invocations, old_recording_interval_started_at_millis

    def empty(self):
        return self.queue.empty() and len(self.invocations[self.FRONT_BUFFER_INDEX]) == 0 and len(
            self.invocations[self.BACK_BUFFER_INDEX]) == 0

    def _update_invocations(self):
        while True:
            try:
                hash_ = self.queue.get()
            except Empty:
                continue
            self.invocations[self.current_index].add(hash_)

    def _toggle_index(self):
        self.recording_interval_started_at_millis = current_milli_time()
        self.current_index = self.FRONT_BUFFER_INDEX \
            if self.current_index == self.BACK_BUFFER_INDEX else self.BACK_BUFFER_INDEX
